@@ -75,9 +75,7 @@ class MyHDFFile(h5py.File):
         h5py.File.__exit__(self, *args)
         if self.remote_file is not None:
             # noinspection PyUnresolvedReferences
-            TransparentPath(self.local_path.name, fs="local").put(
-                self.remote_file
-            )
+            TransparentPath(self.local_path.name, fs="local").put(self.remote_file)
             # noinspection PyUnresolvedReferences
             self.local_path.close()
 
@@ -97,9 +95,7 @@ class MyHDFStore(pd.HDFStore):
     def __exit__(self, exc_type, exc_value, traceback):
         pd.HDFStore.__exit__(self, exc_type, exc_value, traceback)
         if self.remote_file is not None:
-            TransparentPath(self.local_path.name, fs="local").put(
-                self.remote_file
-            )
+            TransparentPath(self.local_path.name, fs="local").put(self.remote_file)
             self.local_path.close()
 
 
@@ -235,7 +231,9 @@ def collapse_ddots(path: Union[Path, TransparentPath, str]) -> TransparentPath:
     )
 
 
-def get_fs(fs_kind: str, project: str, bucket: str) -> Union[gcsfs.GCSFileSystem, LocalFileSystem]:
+def get_fs(
+    fs_kind: str, project: str, bucket: str, token: Optional[Union[str, dict]] = None
+) -> Union[gcsfs.GCSFileSystem, LocalFileSystem]:
     """Gets the FileSystem object of either gcs or local (Default)
 
 
@@ -250,6 +248,9 @@ def get_fs(fs_kind: str, project: str, bucket: str) -> Union[gcsfs.GCSFileSystem
     bucket: str
         bucket name for GCS
 
+    token: Optional[Union[str, dict]]
+        credentials (default value = None)
+
     Returns
     -------
     Union[gcsfs.GCSFileSystem, LocalFileSystem]
@@ -258,21 +259,50 @@ def get_fs(fs_kind: str, project: str, bucket: str) -> Union[gcsfs.GCSFileSystem
     """
     if "gcs" in fs_kind:
         bucket = bucket.replace("/", "")
-        fs = gcsfs.GCSFileSystem(project=project)
+        if token is None:
+            fs = gcsfs.GCSFileSystem(project=project, asynchronous=False)
+        else:
+            fs = gcsfs.GCSFileSystem(project=project, asynchronous=False, token=token)
         # Will raise RefreshError if connection fails
         fs.glob(bucket)
-        if f"{bucket}/" not in fs.buckets:
-            raise NotADirectoryError(
-                f"Bucket {bucket} does not exist in "
-                f"project {project}, or you can not "
-                f"see it. Available buckets are:\n"
-                f"{fs.buckets}"
-            )
+        check_buckets(fs, bucket)
         return fs
     else:
         return LocalFileSystem()
 
 
+def get_buckets(fs: gcsfs.GCSFileSystem) -> List[str]:
+    """Return list of all buckets under the current project."""
+    if "" not in fs.dircache:
+        items = []
+        page = fs.call("GET", "b/", project=fs.project, json_out=True)
+
+        assert page["kind"] == "storage#buckets"
+        items.extend(page.get("items", []))
+        next_page_token = page.get("nextPageToken", None)
+
+        while next_page_token is not None:
+            page = fs.call("GET", "b/", project=fs.project, pageToken=next_page_token, json_out=True,)
+
+            assert page["kind"] == "storage#buckets"
+            items.extend(page.get("items", []))
+            next_page_token = page.get("nextPageToken", None)
+        fs.dircache[""] = [{"name": i["name"] + "/", "size": 0, "type": "directory"} for i in items]
+    return [b["name"] for b in fs.dircache[""]]
+
+
+def check_buckets(fs: Union[gcsfs.GCSFileSystem, LocalFileSystem], bucket: str):
+    if isinstance(fs, gcsfs.GCSFileSystem):
+        buckets = get_buckets(fs)
+        if f"{bucket}/" not in buckets:
+            raise NotADirectoryError(
+                f"Bucket {bucket} does not exist in "
+                f"project {fs.project}, or you can not "
+                f"see it. Available buckets are:\n"
+                f"{buckets}"
+            )
+
+
 class MultipleExistenceError(Exception):
     """Exception raised when a path's destination already contain more than
     one element.
@@ -503,6 +533,8 @@ class TransparentPath(os.PathLike):  # noqa : F811
     unset = True
     cwd = os.getcwd()
     cli = None
+    token = None
+    _do_update_cache = True
 
     _attributes = ["fs", "path", "fs_kind", "project", "bucket"]
 
@@ -531,6 +563,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         project: Optional[str] = None,
         make_main: bool = True,
         nas_dir: Optional[Union[TransparentPath, Path, str]] = None,
+        token: Optional[Union[dict, str]] = None,
     ) -> None:
         """To call before creating any instance to set the file system.
 
@@ -558,6 +591,9 @@ class TransparentPath(os.PathLike):  # noqa : F811
             If specified, TransparentPath will delete any occurence of 'nas_dir' at the beginning of created paths if fs
             is gcs (Default value = None).
 
+        token: Optional[Union[dict, str]]
+            credentials (default value = None)
+
         Returns
         -------
         None
@@ -576,6 +612,9 @@ class TransparentPath(os.PathLike):  # noqa : F811
         if project is not None:
             cls.project = project
 
+        if token is not None:
+            cls.token = token
+
         TransparentPath._set_nas_dir(cls, nas_dir)
 
         if "gcs" in cls.fs_kind:
@@ -585,7 +624,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
                 raise ValueError("Need to provide a project name!")
             cls.fs_kind = f"gcs_{cls.project}"
 
-        cls.fss[cls.fs_kind] = get_fs(cls.fs_kind, cls.project, cls.bucket)
+        cls.fss[cls.fs_kind] = get_fs(cls.fs_kind, cls.project, cls.bucket, cls.token)
         TransparentPath.unset = False
 
         if main_fs is not None:
@@ -977,8 +1016,11 @@ class TransparentPath(os.PathLike):  # noqa : F811
         """
         self.fs.invalidate_cache()
         if "gcs" in self.fs_kind:
-            self.fs.buckets
-            self.fs.info(self.bucket)
+            try:
+                self.fs.info(self.bucket)
+            except FileNotFoundError:
+                # noinspection PyStatementEffect
+                self.buckets
 
     def _cast_fast(self, path: str) -> TransparentPath:
         return TransparentPath(path, fs=self.fs_kind, nocheck=True, bucket=self.bucket, project=self.project)
@@ -988,12 +1030,6 @@ class TransparentPath(os.PathLike):  # noqa : F811
 
     def _set_fs(self) -> None:
         """ Create a new filesystem objet from self, or get the existing one
-
-
-        Returns
-        -------
-        None
-
         """
 
         if "gcs" in self.fs_kind:
@@ -1012,13 +1048,6 @@ class TransparentPath(os.PathLike):  # noqa : F811
         if use_common:
             self.fs = TransparentPath.fss[self.fs_kind]
             self.nas_dir = TransparentPath.nas_dir
-            if "gcs" in self.fs_kind and f"{self.bucket}/" not in self.fs.buckets:
-                raise NotADirectoryError(
-                    f"Bucket {self.bucket} does not exist in "
-                    f"project {self.project}, or you can not "
-                    f"see it. Available buckets are:\n"
-                    f"{self.fs.buckets}"
-                )
         # Else, init a new file system instance and share it with
         # TransparentPath
         else:
@@ -1032,6 +1061,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
                 TransparentPath.bucket = self.bucket
             if TransparentPath.project is None:
                 TransparentPath.project = self.project
+        check_buckets(self.fs, self.bucket)
 
     def _obj_missing(self, obj_name: str, kind: str, *args, **kwargs) -> Any:
         """Method to catch any call to a method/attribute missing from the class.
@@ -1155,7 +1185,8 @@ class TransparentPath(os.PathLike):  # noqa : F811
         """Checks if several objects correspond to the path.
         Raises MultipleExistenceError if so, does nothing if not.
         """
-        self._update_cache()
+        if TransparentPath._do_update_cache:
+            self._update_cache()
         if str(self.__path) == self.bucket or str(self.__path) == "/":
             return
         if not self.exists():
@@ -1280,6 +1311,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
 
         # Asked to remove a directory...
         recursive = kwargs.get("recursive", False)
+
         if recursive:
             if not self.is_dir(exist=True):
                 # ...but self points to something that is not a directory!
@@ -1307,7 +1339,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
             if self.is_dir(exist=True):
                 # Delete anyway
                 if ignore_kind:
-                    self.rm(absent=absent, recursive=True, **kwargs)
+                    self.rm(absent=absent, ignore_kind=True, recursive=True, **kwargs)
                 # or raise
                 else:
                     raise IsADirectoryError("The path points to a directory")
@@ -1419,7 +1451,9 @@ class TransparentPath(os.PathLike):  # noqa : F811
         Iterator[TransparentPath]
 
         """
-        self._update_cache()
+
+        if TransparentPath._do_update_cache:
+            self._update_cache()
         thepath = self / path
         # noinspection PyUnresolvedReferences
         if not thepath.exists():
@@ -1461,7 +1495,8 @@ class TransparentPath(os.PathLike):  # noqa : F811
 
         # Will collapse any '..'
 
-        self._update_cache()
+        if TransparentPath._do_update_cache:
+            self._update_cache()
         if "gcs" in self.fs_kind and str(path) == self.bucket:
             return TransparentPath(fs=self.fs_kind, bucket=self.bucket, project=self.project)
 
@@ -1617,7 +1652,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
             return self.read_text(*args, get_obj=get_obj, update_cache=update_cache, **kwargs)
 
     def read_csv(self, update_cache: bool = True, use_dask: bool = False, **kwargs) -> pd.DataFrame:
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
 
         # noinspection PyTypeChecker,PyUnresolvedReferences
@@ -1645,7 +1680,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
     def read_parquet(
         self, update_cache: bool = True, use_dask: bool = False, **kwargs
     ) -> Union[pd.DataFrame, pd.Series]:
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
 
         index_col, parse_dates, kwargs = get_index_and_date_from_kwargs(**kwargs)
@@ -1672,7 +1707,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
             )
 
     def read_text(self, *args, get_obj: bool = False, update_cache: bool = True, **kwargs) -> Union[str, IO]:
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
         byte_mode = True
         if len(args) == 0:
@@ -1759,7 +1794,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         if use_pandas:
             class_to_use = pd.HDFStore
 
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
         if self.fs_kind == "local":
             # Do not check kwargs since HDFStore and h5py both accepct kwargs anyway
@@ -1774,7 +1809,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         return data
 
     def read_excel(self, update_cache: bool = True, use_dask: bool = False, **kwargs) -> pd.DataFrame:
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
 
         check_kwargs(pd.read_excel, kwargs)
@@ -1928,7 +1963,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
     def write_stuff(
         self, data: Any, *args, overwrite: bool = True, present: str = "ignore", update_cache: bool = True, **kwargs
     ) -> None:
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
         if not overwrite and self.is_file() and present != "ignore":
             raise FileExistsError()
@@ -1965,7 +2000,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         single_file=True, but that is not supported in GCS. If only one file is produced, then this file will be
         moved to self.
         """
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
         if not overwrite and self.is_file() and present != "ignore":
             raise FileExistsError()
@@ -2003,7 +2038,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         Warning : if data is a Dask dataframe, the output will be written in a directory. For convenience, the directory
         if self.with_suffix(""). Reading is transparent and one can specify a path with .parquet suffix.
         """
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
         if not overwrite and self.is_file() and present != "ignore":
             raise FileExistsError()
@@ -2058,7 +2093,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
             mode = kwargs["mode"]
             del kwargs["mode"]
 
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
 
         # If no data is specified, an HDF5 file is returned, opened in write mode, or any other specified mode.
@@ -2140,7 +2175,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         update_cache: bool = True,
         **kwargs,
     ) -> None:
-        if update_cache:
+        if update_cache and TransparentPath._do_update_cache:
             self._update_cache()
         if not overwrite and self.is_file() and present != "ignore":
             raise FileExistsError()
@@ -2196,7 +2231,8 @@ class TransparentPath(os.PathLike):  # noqa : F811
 
         """
 
-        self._update_cache()
+        if TransparentPath._do_update_cache:
+            self._update_cache()
         if present != "raise" and present != "ignore":
             raise ValueError(f"Unexpected value for argument 'present' : {present}")
 
@@ -2307,8 +2343,7 @@ class TransparentPath(os.PathLike):  # noqa : F811
         """used to push a local file to the cloud.
 
         self must be a local TransparentPath. If dst is a TransparentPath, it must be on GCS. If it is a pathlib.Path
-        or a str, it will be casted into a GCS TransparentPath, so a gcs file system must have been set up once
-        before. """
+        or a str, it will be casted into a GCS TransparentPath, so a gcs file system must have been set up before. """
 
         if "gcs" not in "".join(TransparentPath.fss):
             raise ValueError("You need to set up a gcs file system before using the put() command.")
@@ -2328,7 +2363,17 @@ class TransparentPath(os.PathLike):  # noqa : F811
             )
         if type(dst) != TransparentPath:
             dst = TransparentPath(dst, fs="gcs")
-        self.fs.put(self.__fspath__(), dst)
+        if self.is_dir():
+            for item in self.glob("/*"):
+                # noinspection PyUnresolvedReferences
+                item.put(dst / item.name)
+        else:
+            with open(self, "rb") as f1:
+                with open(dst, "wb") as f2:
+                    data = True
+                    while data:
+                        data = f1.read(self.blocksize)
+                        f2.write(data)
 
     def get(self, loc: Union[str, Path, TransparentPath]):
         """used to get a remote file to local.
@@ -2371,5 +2416,10 @@ class TransparentPath(os.PathLike):  # noqa : F811
         return self.exists()
 
     def exists(self):
-        self._update_cache()
+        if TransparentPath._do_update_cache:
+            self._update_cache()
         return self.fs.exists(self.__fspath__())
+
+    @property
+    def buckets(self):
+        return get_buckets(self.fs)
